from .decoder import Decoder, TypeDecoder, decodable, decode
from .encoder import Encoder, TypeEncoder, encodable, encode
from .json_support import JsonDecoder, JsonEncoder, JsonSerializer, json_decode, json_encode
from .mapper import KeyScheme, Mapper
from .serializer import Serializer, serializable

__all__ = [
    "Encoder",
    "JsonEncoder",
    "TypeEncoder",
    "encodable",
    "Decoder",
    "JsonDecoder",
    "TypeDecoder",
    "decodable",
    "decode",
    "serializable",
    "json_encode",
    "json_decode",
    "JsonSerializer",
    "Serializer",
    "Mapper",
    "KeyScheme",
    "encode",
]



from __future__ import annotations

import sys
import typing
from collections import UserString
from dataclasses import MISSING, Field, InitVar, is_dataclass
from enum import Enum
from functools import lru_cache
from inspect import isclass as is_class
from typing import Any, Callable, ClassVar, Dict, List, NewType, Optional, Type, Union, get_type_hints

from chili.error import SerialisationError

try:
    from types import UnionType  # type: ignore

    _SUPPORT_NEW_UNION = True
except ImportError:
    _SUPPORT_NEW_UNION = False
    UnionType = object()


AnnotatedTypeNames = {"AnnotatedMeta", "_AnnotatedAlias"}
_GenericAlias = getattr(typing, "_GenericAlias")
_PROPERTIES = "__typed_properties__"
_DECODE_MAPPER = "__decode_mapper__"
_ENCODE_MAPPER = "__encode_mapper__"
_ENCODABLE = "__encodable__"
_DECODABLE = "__decodable__"
UNDEFINED = object()


__all__ = [
    "get_class_fields",
    "get_dataclass_fields",
    "get_origin_type",
    "get_parameters_map",
    "get_type_args",
    "get_type_hints",
    "get_type_parameters",
    "is_class",
    "is_dataclass",
    "is_enum_type",
    "is_named_tuple",
    "is_optional",
    "is_typed_dict",
    "map_generic_type",
    "unpack_optional",
    "resolve_forward_reference",
    "create_schema",
    "TypeSchema",
    "Property",
]


def get_non_optional_fields(type_name: Type) -> List[str]:
    if is_encodable(type_name):
        schema = getattr(type_name, _ENCODABLE)
    else:
        schema = create_schema(type_name)  # type: ignore

    return [field.name for field in schema.values() if not is_optional(field.type)]


def get_origin_type(type_name: Type) -> Optional[Type]:
    return getattr(type_name, "__origin__", None)


def get_type_args(type_name: Type) -> List[Type]:
    return getattr(type_name, "__args__", [])


def is_encodable(type_name: Type) -> bool:
    return is_class(type_name) and hasattr(type_name, _ENCODABLE)


def is_decodable(type_name: Type) -> bool:
    return is_class(type_name) and hasattr(type_name, _DECODABLE)


def is_serialisable(type_name: Type) -> bool:
    return is_class(type_name) and hasattr(type_name, _DECODABLE) and hasattr(type_name, _ENCODABLE)


def is_optional(type_name: Type) -> bool:
    is_new_union = False
    if _SUPPORT_NEW_UNION:
        is_new_union = type(type_name) is UnionType
    return (
        (get_origin_type(type_name) is Union or type(type_name) is Union or is_new_union)
        and bool(get_type_args(type_name))
        and get_type_args(type_name)[-1] is type(None)  # noqa
    )


def is_newtype(type_name: Type) -> bool:
    if sys.version_info >= (3, 10):
        return isinstance(type_name, NewType)  # type: ignore

    if not hasattr(type_name, "__qualname__"):
        return False

    if type_name.__qualname__ != "NewType.<locals>.new_type":
        return False

    return True


def get_type_parameters(type_name: Type) -> List[Type]:
    return getattr(get_origin_type(type_name), "__parameters__", [])


def get_parameters_map(type_name: Type) -> Dict[Type, Type]:
    args = get_type_args(type_name)
    parameters = get_type_parameters(type_name)

    return dict(zip(parameters, args))


def get_dataclass_fields(type_name: Type) -> Dict[str, Field]:
    return getattr(type_name, "__dataclass_fields__", {})


def get_class_fields(type_name: Type) -> Dict[str, Field]:
    return getattr(type_name, "__annotations__", {})


def unpack_optional(type_name: Type) -> Type:
    return get_type_args(type_name)[0]


def is_enum_type(type_name: Type) -> bool:
    return issubclass(type_name, Enum)


def is_named_tuple(type_name: Type) -> bool:
    return issubclass(type_name, tuple) and hasattr(type_name, "_fields")


def is_typed_dict(type_name: Type) -> bool:
    return issubclass(type_name, dict) and hasattr(type_name, "__annotations__")


def is_user_string(type_name: Type) -> bool:
    return issubclass(type_name, UserString)


def map_generic_type(type_name: Any, type_map: Dict[Any, Any]) -> Any:
    if not type_map:
        return type_name

    type_args = get_type_args(type_name)
    origin_type = get_origin_type(type_name)

    if origin_type is None:  # not a generic type
        return type_name

    if type_args:
        mapped_args = tuple([type_map[arg] for arg in type_args])
        return _GenericAlias(origin_type, mapped_args)

    return type_name


def resolve_forward_reference(module: Any, ref: Union[typing.ForwardRef, str]) -> Any:
    if isinstance(ref, typing.ForwardRef):
        name = ref.__forward_arg__
    else:
        name = ref
    if name in sys.modules[module].__dict__:
        return sys.modules[module].__dict__[name]

    if name in sys.modules["builtins"].__dict__:
        return sys.modules["builtins"].__dict__[name]

    return None


class Property:
    __slots__ = ("name", "type", "_default_value", "_default_factory")

    def __init__(
        self,
        name: str,
        property_type: Type,
        default_value: Any = UNDEFINED,
        default_factory: Callable = None,
    ):
        self.name = name
        self.type = property_type
        self._default_value = default_value if default_value is not MISSING else UNDEFINED
        self._default_factory = default_factory

    @property
    def default_value(self) -> Any:
        if self._default_value is not UNDEFINED:
            return self._default_value

        if self._default_factory:
            return self._default_factory()

        return None

    def __eq__(self, other: Property) -> bool:  # type: ignore
        return self.name == other.name and self.type is other.type


TypeSchema = NewType("TypeSchema", Dict[str, Property])


_default_factories = (list, dict, tuple, set, bytes, bytearray, frozenset)


@lru_cache
def create_schema(cls: Type) -> TypeSchema:
    try:
        properties = typing.get_type_hints(cls, localns=cls.__dict__)  # type: ignore
    except NameError as e:
        raise SerialisationError.invalid_type from e

    schema = TypeSchema({})
    base_classes = [base_class for base_class in cls.__mro__[1:-1] if hasattr(base_class, _PROPERTIES)]
    for base_class in base_classes:
        schema = {**getattr(base_class, _PROPERTIES), **schema}  # type: ignore

    attributes = {name: value for name, value in cls.__dict__.items() if not name.startswith("__")}

    if hasattr(cls, "__dataclass_fields__"):
        for key, value in cls.__dataclass_fields__.items():
            if value.default is not MISSING or value.default_factory is not MISSING:
                attributes[key] = value

    for name, p_type in properties.items():
        p_origin = get_origin_type(p_type)

        if isinstance(p_type, InitVar):
            continue

        # ignore class vars as they are not object properties
        if p_origin and p_origin in (ClassVar, InitVar):
            continue

        if name in attributes:
            prop_value = attributes[name]
            if isinstance(prop_value, Field):
                schema[name] = Property(
                    name,
                    p_type,
                    default_value=prop_value.default,
                    default_factory=prop_value.default_factory,  # type: ignore
                )
            else:
                prop_value_type = type(prop_value)
                if not prop_value and (
                    prop_value_type in _default_factories or isinstance(prop_value, _default_factories)
                ):
                    schema[name] = Property(name, p_type, default_factory=prop_value_type)
                else:
                    schema[name] = Property(name, p_type, default_value=prop_value)
        else:
            schema[name] = Property(name, p_type)

    return schema



from __future__ import annotations

from typing import Any, Generic, Optional, Type, TypeVar

from chili.decoder import Decoder
from chili.encoder import Encoder
from chili.error import SerialisationError
from chili.mapper import Mapper
from chili.typing import (
    _DECODABLE,
    _DECODE_MAPPER,
    _ENCODABLE,
    _ENCODE_MAPPER,
    _PROPERTIES,
    create_schema,
    is_class,
    is_dataclass,
)

C = TypeVar("C")
T = TypeVar("T")


class Serializer(Encoder, Decoder, Generic[T]):
    __generic__: Type[T]

    def __init__(
        self,
        type_encoders: Any = None,
        type_decoders: Any = None,
        encode_mapper: Optional[Mapper] = None,
        decode_mapper: Optional[Mapper] = None,
    ):
        Encoder.__init__(self, type_encoders, mapper=encode_mapper)
        Decoder.__init__(self, type_decoders, mapper=decode_mapper)

    @classmethod
    def __class_getitem__(cls, item: Type[T]) -> Type[Serializer]:  # noqa: E501
        if not is_class(item):
            raise SerialisationError.invalid_type

        if is_dataclass(item):
            item = serializable(item)

        if not hasattr(item, _DECODABLE) and not hasattr(item, _ENCODABLE):
            item = serializable(item)

        return type(  # type: ignore
            f"{cls.__qualname__}[{item.__module__}.{item.__qualname__}]",
            tuple([cls]),
            {
                "__generic__": item,
            },
        )


def serializable(_cls=None, in_mapper: Optional[Mapper] = None, out_mapper: Optional[Mapper] = None) -> Any:
    def _decorate(cls) -> Type[C]:

        setattr(cls, _PROPERTIES, create_schema(cls))
        if in_mapper is not None:
            setattr(cls, _DECODE_MAPPER, in_mapper)
        if out_mapper is not None:
            setattr(cls, _ENCODE_MAPPER, out_mapper)

        setattr(cls, _DECODABLE, True)
        setattr(cls, _ENCODABLE, True)

        return cls

    if _cls is None:
        return _decorate

    return _decorate(_cls)



from json import dumps, loads
from typing import Any, Generic, Type, TypeVar, Union

from .decoder import Decoder, TypeDecoders, decode
from .encoder import Encoder, TypeEncoders, encode
from .serializer import Serializer

T = TypeVar("T")


def json_encode(obj: Any, type_hint: Type = None, type_encoders: TypeEncoders = None) -> str:
    return dumps(encode(obj, type_hint, type_encoders))


def json_decode(json_str: str, type_hint: Type[T], type_decoders: TypeDecoders = None) -> Union[T, Any]:
    return decode(loads(json_str), type_hint, type_decoders)


class JsonEncoder(Encoder, Generic[T]):
    def encode(self, obj: T) -> str:  # type: ignore
        return dumps(super().encode(obj))


class JsonDecoder(Decoder, Generic[T]):
    def decode(self, json_str: str) -> T:  # type: ignore
        return super().decode(loads(json_str))


class JsonSerializer(Serializer, Generic[T]):
    def encode(self, obj: T) -> str:  # type: ignore
        return dumps(super().encode(obj))

    def decode(self, json_str: str) -> T:  # type: ignore
        return super().decode(loads(json_str))



from gaffe import Error


class SerialisationError(Error):
    invalid_generic_type: TypeError
    strict_type_violation: ValueError
    invalid_type: TypeError
    invalid_input: ValueError
    missing_property: KeyError


class DecoderError(SerialisationError):
    ...


class EncoderError(SerialisationError):
    encoding_failed = ...


class MapperError(SerialisationError):
    invalid_schema: ValueError
    invalid_value: ValueError



from collections import namedtuple
from collections.abc import Iterable
from typing import Any, Callable, Dict, Union

from chili.error import MapperError

KeyScheme = namedtuple("KeyScheme", "key scheme")
MappingScheme = Dict[Union[str, Any], Union[Dict, str, Callable, KeyScheme]]


class Mapper:
    def __init__(self, scheme: Union[MappingScheme], preserve_keys: bool = False):
        self.scheme = scheme
        self.preserve_keys = preserve_keys

    def map(self, data: Dict[str, Any], skip_keys: bool = False, default_value: Any = None) -> Dict[str, Any]:
        return self._map(data, self.scheme, skip_keys, default_value)

    def _map(
        self,
        data: Dict,
        scheme: MappingScheme,
        skip_keys: bool,
        default_value: Any = None,
    ) -> Any:
        result = {}
        evaluated_keys = set()
        for new_key, old_key in scheme.items():
            k_type = type(old_key)

            if new_key is Ellipsis:
                map_key = callable(old_key)
                if k_type not in (int, bool) and not map_key:
                    raise MapperError.invalid_schema

                for key, value in data.items():
                    if key in evaluated_keys:
                        continue
                    evaluated_keys.add(key)

                    if map_key:
                        key, value = old_key(key, value)  # type: ignore
                    result[key] = value
                continue
            elif k_type is int or k_type is bool:
                if old_key:
                    evaluated_keys.add(old_key)
                    if skip_keys and old_key not in data:
                        continue
                    result[new_key] = data.get(new_key, default_value)
                continue
            elif k_type is str:
                evaluated_keys.add(old_key)
                if skip_keys and old_key not in data:
                    continue
                result[new_key] = data.get(old_key, default_value)
                continue
            elif k_type is KeyScheme:
                if skip_keys and old_key.key not in data:  # type: ignore
                    continue
                value = data.get(old_key.key, default_value)  # type: ignore
                item_scheme = old_key.scheme  # type: ignore
            elif k_type is dict:
                if skip_keys and new_key not in data:
                    continue
                value = data.get(new_key, default_value)
                item_scheme = old_key
            elif callable(old_key):
                evaluated_keys.add(old_key)
                result[new_key] = old_key(data)
                continue
            else:
                raise MapperError.invalid_schema

            if callable(item_scheme):
                result[new_key] = item_scheme(value)
                continue
            elif isinstance(value, dict):
                result[new_key] = self._map(value, item_scheme, skip_keys, default_value)
                continue
            elif isinstance(value, Iterable):
                result[new_key] = [self._map(item, item_scheme, skip_keys, default_value) for item in iter(value)]
                continue
            else:
                raise MapperError.invalid_value
        if self.preserve_keys:
            return {**{key: value for key, value in data.items() if key not in evaluated_keys}, **result}
        return result



from __future__ import annotations

import collections
import datetime
import decimal
import re
import sys
import typing
from abc import abstractmethod
from base64 import b64encode
from enum import Enum
from functools import lru_cache
from inspect import isclass
from ipaddress import IPv4Address, IPv6Address
from pathlib import Path, PosixPath, PurePath, PurePosixPath, PureWindowsPath, WindowsPath
from typing import Any, Callable, Dict, Generic, List, Pattern, Protocol, Tuple, Type, TypeVar, Union, final
from uuid import UUID

from chili.typing import (
    _ENCODABLE,
    _ENCODE_MAPPER,
    _PROPERTIES,
    UNDEFINED,
    Optional,
    TypeSchema,
    create_schema,
    get_origin_type,
    get_parameters_map,
    get_type_args,
    is_class,
    is_dataclass,
    is_enum_type,
    is_named_tuple,
    is_newtype,
    is_optional,
    is_typed_dict,
    is_user_string,
    map_generic_type,
    resolve_forward_reference,
    unpack_optional,
)

if sys.version_info >= (3, 10):
    from types import UnionType
else:
    UnionType = None

from .error import EncoderError
from .iso_datetime import timedelta_to_iso_duration
from .mapper import Mapper
from .state import StateObject

C = TypeVar("C")
U = TypeVar("U")
T = TypeVar("T")
E = TypeVar("E", bound=Enum)


class TypeEncoder(Protocol):
    @abstractmethod
    def encode(self, value):
        ...


@final
class SimpleEncoder(TypeEncoder, Generic[T]):
    def __init__(self, func: Callable[[Any], T]):
        self._encoder = func

    def encode(self, value: Any) -> T:
        return self._encoder(value)


@final
class ProxyEncoder(Generic[T]):
    def __init__(self, type_annotation: Any) -> None:
        self._encoder = build_type_encoder(type_annotation)

    def decode(self, value: Any) -> T:
        return self._encoder.encode(value)


def encode_regex_to_string(value: Pattern) -> str:
    """
    Encodes regex into string and preserves flags if they are set. Then regex is normally wrapped between slashes.
    """
    flags = []
    if re.I & value.flags:
        flags.append("i")
    if re.M & value.flags:
        flags.append("m")
    if re.S & value.flags:
        flags.append("s")
    if re.X & value.flags:
        flags.append("x")

    if flags:
        return f"/{value.pattern}/{''.join(flags)}"

    return value.pattern


def encodable(_cls=None, mapper: Optional[Mapper] = None) -> Any:
    def _decorate(cls) -> Type[C]:
        # Attach schema to make the class encodable
        if not hasattr(cls, _PROPERTIES):
            setattr(cls, _PROPERTIES, create_schema(cls))
            if mapper:
                setattr(cls, _ENCODE_MAPPER, mapper)

            inner_classes = [
                icls
                for icls in cls.__dict__.values()
                if isclass(icls) and icls.__module__ == cls.__module__ and not hasattr(icls, _PROPERTIES)
            ]
            for inner_class in inner_classes:
                _decorate(inner_class)

        setattr(cls, _ENCODABLE, True)

        return cls

    if _cls is None:
        return _decorate

    return _decorate(_cls)


class TypeEncoders(Dict[Any, TypeEncoder]):
    def __hash__(self) -> int:  # type: ignore[override]
        return hash(tuple(sorted([str(key) for key in self.keys()])))


def ordered_dict(value: collections.OrderedDict) -> List[List[Any]]:
    result = []
    for i_key, i_value in value.items():
        result.append([i_key, i_value])

    return result


_builtin_type_encoders = TypeEncoders(
    {
        bool: SimpleEncoder[bool](bool),
        int: SimpleEncoder[int](int),
        float: SimpleEncoder[float](float),
        str: SimpleEncoder[str](str),
        bytes: SimpleEncoder[str](lambda value: b64encode(value).decode("utf8")),
        bytearray: SimpleEncoder[str](lambda value: b64encode(value).decode("utf8")),
        list: SimpleEncoder[list](list),
        set: SimpleEncoder[list](list),
        frozenset: SimpleEncoder[list](list),
        tuple: SimpleEncoder[list](list),
        dict: SimpleEncoder[dict](dict),
        collections.OrderedDict: SimpleEncoder[list](ordered_dict),
        collections.deque: SimpleEncoder[list](list),
        typing.TypedDict: SimpleEncoder[dict](dict),  # type: ignore
        typing.Dict: SimpleEncoder[dict](dict),
        typing.List: SimpleEncoder[list](list),
        typing.Sequence: SimpleEncoder[list](list),
        typing.Tuple: SimpleEncoder[list](list),  # type: ignore
        typing.Set: SimpleEncoder[list](list),
        typing.FrozenSet: SimpleEncoder[list](list),
        typing.Deque: SimpleEncoder[list](list),
        typing.AnyStr: SimpleEncoder[str](str),  # type: ignore
        decimal.Decimal: SimpleEncoder[str](str),
        datetime.time: SimpleEncoder[str](lambda value: value.isoformat()),
        datetime.date: SimpleEncoder[str](lambda value: value.isoformat()),
        datetime.datetime: SimpleEncoder[str](lambda value: value.isoformat()),
        datetime.timedelta: SimpleEncoder[str](timedelta_to_iso_duration),
        PurePath: SimpleEncoder[str](str),
        PureWindowsPath: SimpleEncoder[str](str),
        PurePosixPath: SimpleEncoder[str](str),
        Path: SimpleEncoder[str](str),
        PosixPath: SimpleEncoder[str](str),
        WindowsPath: SimpleEncoder[str](str),
        Pattern: SimpleEncoder[str](encode_regex_to_string),
        re.Pattern: SimpleEncoder[str](encode_regex_to_string),
        IPv6Address: SimpleEncoder[str](str),
        IPv4Address: SimpleEncoder[str](str),
        UUID: SimpleEncoder[str](str),
    }
)


class ListEncoder(TypeEncoder):
    def __init__(self, item_encoder: TypeEncoder):
        self.item_encoder = item_encoder

    def encode(self, value: typing.List) -> list:
        return [self.item_encoder.encode(item) for item in value]


class TupleEncoder(TypeEncoder):
    def __init__(self, items_encoder: List[Union[TypeEncoder, Any]]):
        self.items_encoder = items_encoder
        self._variadic = items_encoder[-1] is Ellipsis

    def encode(self, value: Tuple) -> List:
        if self._variadic:
            return self._encode_variadic(value)
        return self._encode_static(value)

    def _encode_variadic(self, value: Tuple) -> List:
        result = []
        type_encoders = self.items_encoder[:-1]
        num_encoders = len(type_encoders)

        for index, item in enumerate(value):
            encoder = type_encoders[min(index, num_encoders - 1)]
            result.append(encoder.encode(item))

        return result

    def _encode_static(self, value: Tuple) -> List:
        result = [encoder.encode(value[index]) for index, encoder in enumerate(self.items_encoder)]
        return result


class DictEncoder(TypeEncoder):
    def __init__(self, encoders: List[TypeEncoder]):
        self.key_encoder, self.value_encoder = encoders

    def encode(self, value: dict) -> dict:
        return {self.key_encoder.encode(i_key): self.value_encoder.encode(i_value) for i_key, i_value in value.items()}


class OrderedDictEncoder(TypeEncoder):
    def __init__(self, encoders: List[TypeEncoder]):
        self.key_encoder, self.value_encoder = encoders

    def encode(self, value: dict) -> list:
        return [
            [self.key_encoder.encode(i_key), self.value_encoder.encode(i_value)] for i_key, i_value in value.items()
        ]


class ClassEncoder(TypeEncoder):
    _fields: Dict[str, TypeEncoder]
    _schema: TypeSchema

    def __init__(self, class_name: Type, extra_encoders: TypeEncoders = None, force: bool = False):
        self.class_name = class_name
        self._extra_encoders = extra_encoders
        self._schema = create_schema(class_name)  # type: ignore
        self.force = force

    def encode(self, value: Any) -> StateObject:
        if not isinstance(value, self.class_name):
            raise EncoderError.invalid_input

        if not hasattr(self, "_fields"):
            self._fields = self._build()

        result = {}
        for name, field in self._schema.items():
            prop_value = self._fields[name].encode(getattr(value, name, field.default_value))
            if prop_value is not UNDEFINED:
                result[name] = prop_value

        return result

    def _build(self) -> Dict[str, TypeEncoder]:
        return {name: self._build_type_encoder(field.type) for name, field in self._schema.items()}

    def _build_type_encoder(self, a_type: Type) -> TypeEncoder:
        return build_type_encoder(a_type, self._extra_encoders, self.class_name.__module__, self.force)  # type: ignore


class GenericClassEncoder(ClassEncoder):
    def __init__(self, class_name: Type, extra_encoders: TypeEncoders = None, force: bool = False):
        self._generic_type = class_name
        self._extra_encoders = extra_encoders
        self._generic_parameters = get_parameters_map(class_name)
        self.force = force
        type_: Type = get_origin_type(class_name)  # type: ignore
        super().__init__(type_)

    def _build_type_encoder(self, a_type: Type) -> TypeEncoder:
        return build_type_encoder(
            map_generic_type(a_type, self._generic_parameters),
            self._extra_encoders,
            self._generic_type.__module__,
            self.force,
        )


class EnumEncoder(TypeEncoder, Generic[E]):
    def __init__(self, enum_type: Type):
        self.enum_type = enum_type

    def encode(self, value: E) -> Any:
        if not isinstance(value, self.enum_type):
            raise EncoderError.invalid_input

        return value.value


class NamedTupleEncoder(TypeEncoder):
    def __init__(self, class_name: Type, extra_encoders: TypeEncoders = None, force: bool = False):
        self.type = class_name
        self._is_typed = hasattr(class_name, "__annotations__")
        self._arg_encoders: List[TypeEncoder] = []
        self._extra_encoders = extra_encoders
        self.force = force
        if self._is_typed:
            self._build()

    def encode(self, value: tuple) -> list:
        if not self._is_typed:
            return list(value)

        return [
            self._arg_encoders[index].encode(item) if index < len(self._arg_encoders) else item
            for index, item in enumerate(value)
        ]

    def _build(self) -> None:
        field_types = self.type.__annotations__
        for item_type in field_types.values():
            self._arg_encoders.append(
                build_type_encoder(item_type, self._extra_encoders, self.type.__module__, self.force)
            )


class TypedDictEncoder(TypeEncoder):
    def __init__(self, class_name: Type, extra_encoders: TypeEncoders = None, force: bool = False):
        self.type = class_name
        self._key_encoders = {}
        self.force = force
        for key_name, key_type in class_name.__annotations__.items():
            self._key_encoders[key_name] = build_type_encoder(
                key_type, extra_encoders, class_name.__module__, self.force
            )

    def encode(self, value: dict) -> dict:
        return {key: self._key_encoders[key].encode(item) for key, item in value.items()}


class OptionalTypeEncoder(TypeEncoder):
    def __init__(self, type_encoder: TypeEncoder):
        self._encoder = type_encoder

    def encode(self, value: Any) -> Any:
        if value is None:
            return None

        return self._encoder.encode(value)


class UnionEncoder(TypeEncoder):
    def __init__(self, supported_types: List[Type], extra_encoders: TypeEncoders = None, force: bool = False):
        self.supported_types = supported_types
        self._extra_encoders = extra_encoders
        self.force = force

    def encode(self, value: Any) -> Any:
        value_type = type(value)
        if value_type not in self.supported_types:
            raise EncoderError.invalid_input

        return build_type_encoder(value_type, self._extra_encoders, force=self.force).encode(value)  # type: ignore


_supported_generics = {
    list: ListEncoder,
    tuple: TupleEncoder,
    dict: DictEncoder,
    set: ListEncoder,
    frozenset: ListEncoder,
    collections.deque: ListEncoder,
    collections.OrderedDict: OrderedDictEncoder,
}


@lru_cache(maxsize=None)
def build_type_encoder(
    a_type: Type, extra_encoders: TypeEncoders = None, module: Any = None, force: bool = False
) -> TypeEncoder:
    if extra_encoders and a_type in extra_encoders:
        return extra_encoders[a_type]

    if a_type in _builtin_type_encoders:
        return _builtin_type_encoders[a_type]

    origin_type = get_origin_type(a_type)

    if origin_type is None and is_dataclass(a_type):
        if issubclass(a_type, Generic):  # type: ignore
            raise EncoderError.invalid_type
        return ClassEncoder(a_type, extra_encoders, force)

    if origin_type and is_dataclass(origin_type):
        if issubclass(origin_type, Generic):  # type: ignore
            return GenericClassEncoder(a_type)
        return ClassEncoder(a_type, extra_encoders, force)

    if origin_type is None:
        origin_type = a_type

    if is_class(origin_type) and is_enum_type(origin_type):
        return EnumEncoder(origin_type)

    if is_class(origin_type) and is_named_tuple(origin_type):
        return NamedTupleEncoder(origin_type, extra_encoders, force)

    if is_class(origin_type) and is_typed_dict(origin_type):
        return TypedDictEncoder(origin_type, extra_encoders, force)

    if is_class(origin_type) and is_user_string(origin_type):
        return SimpleEncoder[str](str)

    if origin_type is Union or (UnionType and isinstance(origin_type, UnionType)):
        type_args = get_type_args(a_type)
        if len(type_args) == 2 and type_args[-1] is type(None):
            return OptionalTypeEncoder(build_type_encoder(type_args[0], extra_encoders, module, force))  # type: ignore
        return UnionEncoder(type_args, extra_encoders, force=force)

    if isinstance(a_type, typing.ForwardRef) and module is not None:
        resolved_reference = resolve_forward_reference(module, a_type)
        if resolved_reference is not None:
            return Encoder[resolved_reference](encoders=extra_encoders)  # type: ignore[valid-type]

    if typing.get_origin(origin_type) is not None:
        raise EncoderError.invalid_type(a_type)

    if hasattr(origin_type, _PROPERTIES):
        if origin_type != a_type:
            return GenericClassEncoder(a_type)
        return Encoder[origin_type](encoders=extra_encoders)  # type: ignore[valid-type]

    if isinstance(a_type, TypeVar):
        if a_type.__bound__ is None:
            raise EncoderError.invalid_type(a_type)
        return build_type_encoder(a_type.__bound__, extra_encoders, module, force)

    if is_newtype(a_type):
        return build_type_encoder(a_type.__supertype__, extra_encoders, module, force)

    if origin_type not in _supported_generics:
        if is_class(origin_type) and force:
            return Encoder[origin_type](encoders=extra_encoders)  # type: ignore[valid-type]
        raise EncoderError.invalid_type(type=a_type)

    type_attributes: List[TypeEncoder] = [
        build_type_encoder(subtype, extra_encoders=extra_encoders, module=module, force=force)  # type: ignore
        if subtype is not ...
        else ...  # noqa: E501
        for subtype in get_type_args(a_type)
    ]

    if len(type_attributes) == 1:
        return _supported_generics[origin_type](type_attributes[0])  # type: ignore

    return _supported_generics[origin_type](type_attributes)  # type: ignore


class Encoder(Generic[T]):
    __generic__: Type[T]
    _encoders: Dict[str, TypeEncoder]

    def __init__(self, encoders: Union[Dict, TypeEncoders] = None, mapper: Optional[Mapper] = None):
        if encoders and not isinstance(encoders, TypeEncoders):
            encoders = TypeEncoders(encoders)
        self.encode_mapper = mapper
        self.type_encoders = encoders

    def encode(self, obj: T) -> StateObject:
        if not hasattr(self, "_encoders"):
            self._encoders = self._build_encoders()

        result = {}
        for key, prop in self.schema.items():
            if hasattr(obj, key):
                value = getattr(obj, key)
            elif is_optional(prop.type):
                value = prop.default_value
            else:
                continue
            result[key] = self._encoders[prop.name].encode(value)

        if hasattr(self.__generic__, _ENCODE_MAPPER):
            mapper = getattr(self.__generic__, _ENCODE_MAPPER)
            return mapper.map(result)
        elif self.encode_mapper:
            return self.encode_mapper.map(result)

        return result

    @property
    def schema(self) -> TypeSchema:
        return getattr(self.__generic__, _PROPERTIES)

    def _build_encoders(self) -> Dict[str, TypeEncoder]:
        schema: TypeSchema = self.schema

        return {
            prop.name: build_type_encoder(prop.type, extra_encoders=self.type_encoders, force=True)  # type: ignore
            for prop in schema.values()
        }

    @classmethod
    def __class_getitem__(cls, item: Any) -> Type[Encoder]:  # noqa: E501
        if not isclass(item):
            raise EncoderError.invalid_generic_type

        if is_dataclass(item):
            item = encodable(item)

        if not hasattr(item, _ENCODABLE):
            item = encodable(item)

        return type(  # type: ignore
            f"{cls.__qualname__}[{item.__module__}.{item.__qualname__}]",
            tuple([cls]),
            {
                "__generic__": item,
            },
        )


def encode(
    obj: Any,
    type_hint: Type = None,
    encoders: Union[TypeEncoders, Dict[Any, TypeEncoder]] = None,
    force: bool = False,
) -> StateObject:
    if encoders and not isinstance(encoders, TypeEncoders):
        encoders = TypeEncoders(encoders)

    if type_hint is not None:
        encoder = build_type_encoder(type_hint, extra_encoders=encoders, force=force)  # type: ignore
    else:
        encoder = build_type_encoder(type(obj), extra_encoders=encoders, force=force)  # type: ignore

    if encoder is None:
        raise EncoderError.invalid_input

    return encoder.encode(obj)



from __future__ import annotations

import collections
import datetime
import decimal
import re
import sys
import typing
from abc import abstractmethod
from base64 import b64decode
from enum import Enum
from functools import lru_cache
from inspect import isclass
from ipaddress import IPv4Address, IPv6Address
from pathlib import Path, PosixPath, PurePath, PurePosixPath, PureWindowsPath, WindowsPath
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    List,
    Optional,
    Pattern,
    Protocol,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
    final,
    get_origin,
    get_type_hints,
)
from uuid import UUID

from chili.typing import (
    _DECODABLE,
    _DECODE_MAPPER,
    _PROPERTIES,
    UNDEFINED,
    TypeSchema,
    create_schema,
    get_non_optional_fields,
    get_origin_type,
    get_parameters_map,
    get_type_args,
    is_class,
    is_dataclass,
    is_decodable,
    is_enum_type,
    is_named_tuple,
    is_newtype,
    is_optional,
    is_typed_dict,
    is_user_string,
    map_generic_type,
    resolve_forward_reference,
    unpack_optional,
)

if sys.version_info >= (3, 10):
    from types import UnionType
else:
    UnionType = None

from .error import DecoderError
from .iso_datetime import parse_iso_date, parse_iso_datetime, parse_iso_duration, parse_iso_time
from .mapper import Mapper
from .state import StateObject

C = TypeVar("C")
U = TypeVar("U")
T = TypeVar("T")
E = TypeVar("E", bound=Enum)


class TypeDecoder(Protocol):
    @abstractmethod
    def decode(self, value):
        ...


@final
class SimpleDecoder(Generic[T]):
    def __init__(self, func: Callable[[Any], T]) -> None:
        self._decoder = func

    def decode(self, value: Any) -> T:
        return self._decoder(value)


@final
class ProxyDecoder(Generic[T]):
    def __init__(self, type_annotation: Any) -> None:
        self._decoder = build_type_decoder(type_annotation)

    def decode(self, value: Any) -> T:
        return self._decoder.decode(value)


_REGEX_FLAGS = {
    "i": re.I,
    "m": re.M,
    "s": re.S,
    "x": re.X,
}


def decode_regex_from_string(value: str) -> Pattern:
    """
    Decodes regex from string containing flags /ims.
    """
    if value.startswith("/"):
        if value.endswith("/"):
            return re.compile(value[1:-1])
        elif value[-2] == "/":
            pattern = value[1:-2]
            flags = value[-1]
        elif value[-3] == "/":
            pattern = value[1:-3]
            flags = value[-2:]
        elif value[-4] == "/":
            pattern = value[1:-4]
            flags = value[-3:]
        elif value[-5] == "/":
            pattern = value[1:-5]
            flags = value[-4:]
        else:
            raise DecoderError.invalid_input(value)
    else:
        return re.compile(value)

    return re.compile(pattern, flags=sum(_REGEX_FLAGS[flag] for flag in flags))


def decodable(_cls=None, mapper: Optional[Mapper] = None) -> Any:
    def _decorate(cls) -> Type[C]:
        # Attach schema to make the class decodable
        if not hasattr(cls, _PROPERTIES):
            setattr(cls, _PROPERTIES, create_schema(cls))
            if mapper:
                setattr(cls, _DECODE_MAPPER, mapper)

            inner_classes = [
                icls
                for icls in cls.__dict__.values()
                if isclass(icls) and icls.__module__ == cls.__module__ and not hasattr(icls, _PROPERTIES)
            ]
            for inner_class in inner_classes:
                _decorate(inner_class)

        setattr(cls, _DECODABLE, True)

        return cls

    if _cls is None:
        return _decorate

    return _decorate(_cls)


class TypeDecoders(Dict[Any, TypeDecoder]):
    def __hash__(self) -> int:  # type: ignore
        return hash(tuple(sorted([str(key) for key in self.keys()])))


def ordered_dict(value: List[List[Any]]) -> collections.OrderedDict:
    result = collections.OrderedDict()
    for item in value:
        result[item[0]] = item[1]

    return result


_builtin_type_decoders = TypeDecoders(
    {
        bool: SimpleDecoder[bool](bool),
        int: SimpleDecoder[int](int),
        float: SimpleDecoder[float](float),
        str: SimpleDecoder[str](str),
        bytes: SimpleDecoder[bytes](lambda value: b64decode(value.encode("utf8"))),
        bytearray: SimpleDecoder[bytearray](lambda value: bytearray(b64decode(value.encode("utf8")))),
        list: SimpleDecoder[list](list),
        set: SimpleDecoder[set](set),
        frozenset: SimpleDecoder[frozenset](frozenset),
        tuple: SimpleDecoder[tuple](tuple),
        dict: SimpleDecoder[dict](dict),
        collections.OrderedDict: SimpleDecoder[collections.OrderedDict](ordered_dict),
        collections.deque: SimpleDecoder[collections.deque](collections.deque),
        typing.TypedDict: SimpleDecoder[typing.TypedDict](typing.TypedDict),  # type: ignore
        typing.Dict: SimpleDecoder[dict](dict),
        typing.List: SimpleDecoder[list](list),
        typing.Sequence: SimpleDecoder[list](list),
        typing.Tuple: SimpleDecoder[tuple](tuple),  # type: ignore
        typing.Set: SimpleDecoder[set](set),
        typing.FrozenSet: SimpleDecoder[frozenset](frozenset),
        typing.Deque: SimpleDecoder[typing.Deque](typing.Deque),
        typing.AnyStr: SimpleDecoder[str](str),  # type: ignore
        decimal.Decimal: SimpleDecoder[decimal.Decimal](decimal.Decimal),
        datetime.time: SimpleDecoder[datetime.time](parse_iso_time),
        datetime.date: SimpleDecoder[datetime.date](parse_iso_date),
        datetime.datetime: SimpleDecoder[datetime.datetime](parse_iso_datetime),
        datetime.timedelta: SimpleDecoder[datetime.timedelta](parse_iso_duration),
        PurePath: SimpleDecoder[PurePath](PurePath),
        PureWindowsPath: SimpleDecoder[PureWindowsPath](PureWindowsPath),
        PurePosixPath: SimpleDecoder[PurePosixPath](PurePosixPath),
        Path: SimpleDecoder[Path](Path),
        PosixPath: SimpleDecoder[PosixPath](PosixPath),
        WindowsPath: SimpleDecoder[WindowsPath](WindowsPath),
        Pattern: SimpleDecoder[Pattern](decode_regex_from_string),
        re.Pattern: SimpleDecoder[re.Pattern](decode_regex_from_string),
        IPv4Address: SimpleDecoder[IPv4Address](IPv4Address),
        IPv6Address: SimpleDecoder[IPv6Address](IPv6Address),
        UUID: SimpleDecoder[UUID](UUID),
    }
)


class ListDecoder(TypeDecoder):
    def __init__(self, item_decoder: TypeDecoder):
        self.item_decoder = item_decoder

    def decode(self, value: list) -> list:
        res = []
        for item in value:
            decoded = self.item_decoder.decode(item)
            res.append(decoded)

        return res


class TupleDecoder(TypeDecoder):
    def __init__(self, items_decoder: List[Union[TypeDecoder, Any]]):
        self.items_decoder = items_decoder
        self._variadic = items_decoder[-1] is Ellipsis

    def decode(self, value: Sequence) -> Tuple:
        if self._variadic:
            return self._decode_variadic(value)
        return self._decode_static(value)

    def _decode_variadic(self, value: Sequence) -> Tuple:
        result = []
        type_decoders = self.items_decoder[:-1]
        num_decoders = len(type_decoders)

        for index, item in enumerate(value):
            decoder = type_decoders[min(index, num_decoders - 1)]
            result.append(decoder.decode(item))

        return tuple(result)

    def _decode_static(self, value: Sequence) -> Tuple:
        result = [decoder.decode(value[index]) for index, decoder in enumerate(self.items_decoder)]
        return tuple(result)


class DictDecoder(TypeDecoder):
    def __init__(self, encoders: List[TypeDecoder]) -> None:
        self.key_decoder, self.value_decoder = encoders

    def decode(self, value: dict) -> dict:
        return {self.key_decoder.decode(i_key): self.value_decoder.decode(i_value) for i_key, i_value in value.items()}


class OrderedDictDecoder(TypeDecoder):
    def __init__(self, decoders: List[TypeDecoder]):
        self.key_decoder, self.value_decoder = decoders

    def decode(self, value: list) -> collections.OrderedDict:
        result = collections.OrderedDict()

        for item in value:
            result[self.key_decoder.decode(item[0])] = self.value_decoder.decode(item[1])

        return result


class UnionDecoder(TypeDecoder):
    _PRIMITIVE_TYPES = {int, float, bool, str}
    _TRANSFORMABLE_TYPES = {
        datetime.time,
        datetime.datetime,
        datetime.timedelta,
        bytes,
        bytearray,
    }
    _CASTABLES_TYPES = {decimal.Decimal}

    def __init__(self, valid_types: List[Type], extra_decoders: TypeDecoders = None, force: bool = False):
        self.valid_types = valid_types
        self._type_decoders = {}

        for a_type in valid_types:
            if a_type in self._PRIMITIVE_TYPES:
                self._type_decoders[a_type] = a_type
                continue
            self._type_decoders[a_type] = build_type_decoder(
                a_type, extra_decoders=extra_decoders, force=force  # type: ignore
            )

        self.force = force

    def decode(self, value: Any) -> Any:
        passed_type = type(value)

        if passed_type in self._PRIMITIVE_TYPES:
            return self._type_decoders.get(passed_type, str)(value)

        if passed_type in {int, str}:
            for castable, decoder in self._type_decoders.items():
                try:
                    return castable(value)
                except Exception:
                    continue

        if passed_type is str:
            for transformable, decoder in self._type_decoders.items():
                try:
                    return decoder.decode(value)
                except Exception:
                    continue

        if passed_type is dict:
            provided_fields = set(value.keys())
            if self.force:
                decoders = self._type_decoders
            else:
                decoders = {
                    class_name: decoder
                    for class_name, decoder in self._type_decoders.items()
                    if is_decodable(class_name) or is_dataclass(class_name)
                }
            # Greedy matching
            for class_name, decoder in decoders.items():
                expected_fields = set(get_non_optional_fields(class_name))
                if not provided_fields.issubset(expected_fields):
                    continue
                try:
                    return decoder.decode(value)
                except Exception:
                    continue
            # Non-greedy matching
            for class_name, decoder in decoders.items():
                expected_fields = set(get_non_optional_fields(class_name))
                if not expected_fields.issubset(provided_fields):
                    continue
                try:
                    return decoder.decode(value)
                except Exception:
                    continue

        raise DecoderError.invalid_input(value)


class ClassDecoder(TypeDecoder):
    _fields: Dict[str, TypeDecoder]
    _schema: TypeSchema

    def __init__(self, class_name: Type, extra_decoders: TypeDecoders = None, force: bool = False):
        self.class_name = class_name
        self._schema = create_schema(class_name)  # type: ignore
        self._extra_decoders = extra_decoders
        self.force = force

    def decode(self, value: StateObject) -> Any:
        if not isinstance(value, dict):
            raise DecoderError.invalid_input

        if not hasattr(self, "_fields"):
            self._fields = self._build()

        instance = self.class_name.__new__(self.class_name)  # type: ignore

        for key, prop in self._schema.items():
            prop_value = (
                self._fields[key].decode(value[key]) if key in value else getattr(prop, "default_value", UNDEFINED)
            )
            if prop_value is not UNDEFINED:
                setattr(instance, key, prop_value)

        if hasattr(instance, "__post_init__"):
            instance.__post_init__()

        return instance

    def _build(self) -> Dict[str, TypeDecoder]:
        return {name: self._build_type_decoder(field.type) for name, field in self._schema.items()}

    def _build_type_decoder(self, a_type: Type) -> TypeDecoder:
        return build_type_decoder(a_type, self._extra_decoders, self.class_name.__module__, self.force)  # type: ignore


class GenericClassDecoder(ClassDecoder):
    def __init__(self, class_name: Type, extra_decoders: TypeDecoders = None, force: bool = False):
        self._generic_type = class_name
        self._generic_parameters = get_parameters_map(class_name)
        self._extra_decoders = extra_decoders
        type_: Type = get_origin_type(class_name)  # type: ignore
        self.force = force
        super().__init__(type_)

    def _build_type_decoder(self, a_type: Type) -> TypeDecoder:
        return build_type_decoder(
            map_generic_type(a_type, self._generic_parameters),
            self._extra_decoders,
            self._generic_type.__module__,
            self.force,
        )


class EnumDecoder(TypeDecoder, Generic[E]):
    def __init__(self, class_name: Type):
        self.class_name = class_name

    def decode(self, value: Any) -> E:
        return self.class_name(value)


class NamedTupleDecoder(TypeDecoder):
    def __init__(self, class_name: Type, extra_decoders: TypeDecoders = None, force: bool = False):
        self.class_name = class_name
        self._is_typed = hasattr(class_name, "__annotations__")
        self._arg_decoders: List[TypeDecoder] = []
        self._extra_decoders = extra_decoders
        self.force = force
        if self._is_typed:
            self._build()

    def decode(self, value: list) -> tuple:
        if not self._is_typed:
            return tuple(value)

        decoded_values = []
        for index, item in enumerate(value):
            if index < len(self._arg_decoders):
                decoded_values.append(self._arg_decoders[index].decode(item))
                continue

            decoded_values.append(item)

        return tuple(decoded_values)

    def _build(self) -> None:
        field_types = self.class_name.__annotations__
        for item_type in field_types.values():
            self._arg_decoders.append(
                build_type_decoder(item_type, self._extra_decoders, self.class_name.__module__, self.force)
            )


class TypedDictDecoder(TypeDecoder):
    def __init__(self, class_name: Type, extra_decoders: TypeDecoders = None, force: bool = False):
        self.class_name = class_name
        self._key_decoders = {}
        self._extra_decoders = extra_decoders
        for key_name, key_type in class_name.__annotations__.items():
            self._key_decoders[key_name] = build_type_decoder(
                key_type, self._extra_decoders, class_name.__module__, force
            )

    def decode(self, value: dict) -> dict:
        return {key: self._key_decoders[key].decode(item) for key, item in value.items()}


class OptionalTypeDecoder(TypeDecoder):
    def __init__(self, type_decoder: TypeDecoder):
        self._decoder = type_decoder

    def decode(self, value: Any) -> Any:
        if value is None:
            return None

        return self._decoder.decode(value)


_supported_generics = {
    list: ListDecoder,
    tuple: TupleDecoder,
    dict: DictDecoder,
    set: ListDecoder,
    frozenset: ListDecoder,
    collections.deque: ListDecoder,
    collections.OrderedDict: OrderedDictDecoder,
}


@lru_cache(maxsize=None)
def build_type_decoder(
    a_type: Type, extra_decoders: TypeDecoders = None, module: Any = None, force: bool = False
) -> TypeDecoder:
    if extra_decoders and a_type in extra_decoders:
        return extra_decoders[a_type]

    if a_type in _builtin_type_decoders:
        return _builtin_type_decoders[a_type]

    origin_type = get_origin_type(a_type)

    if origin_type is None and is_dataclass(a_type):
        if issubclass(a_type, Generic):  # type: ignore
            raise DecoderError.invalid_type
        return ClassDecoder(a_type, extra_decoders, force)

    if origin_type and is_dataclass(origin_type):
        if issubclass(origin_type, Generic):  # type: ignore
            return GenericClassDecoder(a_type)
        return ClassDecoder(a_type, extra_decoders, force)

    if origin_type is None:
        origin_type = a_type

    if is_class(origin_type) and is_enum_type(origin_type):
        return EnumDecoder(origin_type)

    if is_class(origin_type) and is_named_tuple(origin_type):
        return NamedTupleDecoder(origin_type, extra_decoders)

    if is_class(origin_type) and is_typed_dict(origin_type):
        return TypedDictDecoder(origin_type, extra_decoders, force)

    if is_class(origin_type) and is_user_string(origin_type):
        return SimpleDecoder[origin_type](origin_type)  # type: ignore

    if origin_type is Union or (UnionType and isinstance(origin_type, UnionType)):
        type_args = get_type_args(a_type)
        if len(type_args) == 2 and type_args[-1] is type(None):  # type: ignore
            return OptionalTypeDecoder(
                build_type_decoder(a_type=type_args[0], extra_decoders=extra_decoders, force=force)  # type: ignore
            )
        return UnionDecoder(type_args, extra_decoders=extra_decoders, force=force)

    if isinstance(a_type, typing.ForwardRef) and module is not None:
        resolved_reference = resolve_forward_reference(module, a_type)
        if resolved_reference is not None:
            return Decoder[resolved_reference](decoders=extra_decoders)  # type: ignore

    if isinstance(a_type, TypeVar):
        if a_type.__bound__ is None:
            raise DecoderError.invalid_type(a_type)
        return build_type_decoder(a_type.__bound__, extra_decoders, module, force)

    if is_newtype(a_type):
        return build_type_decoder(a_type.__supertype__, extra_decoders, module, force)

    if get_origin(origin_type) is not None:
        raise DecoderError.invalid_type(a_type)

    if hasattr(origin_type, _PROPERTIES):
        return Decoder[origin_type](decoders=extra_decoders)  # type: ignore

    if is_optional(a_type):
        return OptionalTypeDecoder(
            build_type_decoder(unpack_optional(a_type), extra_decoders, module, force)  # type: ignore
        )

    if origin_type not in _supported_generics:
        if force and is_class(origin_type):
            return Decoder[origin_type](extra_decoders)  # type: ignore
        raise DecoderError.invalid_type(a_type)

    type_attributes: List[Union[TypeDecoder, Any]] = [
        build_type_decoder(subtype, extra_decoders=extra_decoders, module=module)  # type: ignore
        if subtype is not ...
        else ...
        for subtype in get_type_args(a_type)
    ]
    if len(type_attributes) == 1:
        return _supported_generics[origin_type](type_attributes[0])  # type: ignore

    return _supported_generics[origin_type](type_attributes)  # type: ignore


class Decoder(Generic[T]):
    __generic__: Type[T]
    _decoders: Dict[str, TypeDecoder]

    def __init__(self, decoders: Union[TypeDecoders, Dict[Any, TypeDecoder]] = None, mapper: Optional[Mapper] = None):
        if decoders and not isinstance(decoders, TypeDecoders):
            decoders = TypeDecoders(decoders)
        self.decode_mapper = mapper
        self.type_decoders = decoders

    @property
    def schema(self) -> TypeSchema:
        return getattr(self.__generic__, _PROPERTIES)

    def decode(self, obj: Dict[str, StateObject]) -> T:
        if not hasattr(self, "_decoders"):
            self._decoders = self._build_decoders()

        instance = self.__generic__.__new__(self.__generic__)

        if hasattr(self.__generic__, _DECODE_MAPPER):
            mapper = getattr(self.__generic__, _DECODE_MAPPER)
            obj = mapper.map(obj)
        elif self.decode_mapper:
            obj = self.decode_mapper.map(obj)

        for key, prop in self.schema.items():
            if key not in obj:
                if is_optional(prop.type):
                    value = prop.default_value
                else:
                    raise DecoderError.missing_property(key=key)
            else:
                value = self._decoders[prop.name].decode(obj[key])

            try:
                setattr(instance, prop.name, value)
            except AttributeError:
                setattr(instance, f"_{prop.name}", value)

        return instance

    def _build_decoders(self) -> Dict[str, TypeDecoder]:
        schema: TypeSchema = getattr(self.__generic__, _PROPERTIES)

        return {
            prop.name: build_type_decoder(prop.type, extra_decoders=self.type_decoders, force=True)  # type: ignore
            for prop in schema.values()
        }

    @classmethod
    def __class_getitem__(cls, item: Type) -> Type[Decoder]:  # noqa: E501
        if not isclass(item):
            raise DecoderError.invalid_generic_type

        if is_dataclass(item):
            item = decodable(item)

        if not hasattr(item, _DECODABLE):
            item = decodable(item)

        return type(  # type: ignore
            f"{cls.__qualname__}[{item.__module__}.{item.__qualname__}]",
            tuple([cls]),
            {
                "__generic__": item,
            },
        )


def decode(
    obj: StateObject, a_type: Type[T], decoders: Union[TypeDecoders, Dict[Any, TypeDecoder]] = None, force: bool = False
) -> T:
    if decoders and not isinstance(decoders, TypeDecoders):
        decoders = TypeDecoders(decoders)

    decoder = build_type_decoder(a_type, extra_decoders=decoders, force=force)  # type: ignore
    if decoder is None:
        raise DecoderError.invalid_type

    return decoder.decode(obj)



from __future__ import annotations

from typing import Dict, Union

StatefulTypes = Union[list, str, int, float, bool, Dict[str, "StatefulTypes"], tuple]  # type: ignore

StateObject = Union[None, StatefulTypes, Dict[str, StatefulTypes]]  # type: ignore[misc]



import re
from datetime import date, datetime, time, timedelta, timezone

__all__ = [
    "parse_iso_datetime",
    "parse_iso_date",
    "parse_iso_duration",
    "parse_iso_time",
    "timedelta_to_iso_duration",
]


ISO_8601_DATETIME_REGEX = re.compile(
    r"^(\d{4})-?([0-1]\d)-?([0-3]\d)[t\s]?([0-2]\d:?[0-5]\d:?[0-5]\d|23:59:60|235960)(\.\d+)?(z|[+-]\d{2}:\d{2})?$",
    re.I,
)
ISO_8601_DATE_REGEX = re.compile(r"^(\d{4})-?([0-1]\d)-?([0-3]\d)$", re.I)
ISO_8601_TIME_REGEX = re.compile(
    r"^(?P<time>[0-2]\d:?[0-5]\d:?[0-5]\d|23:59:60|235960)(?P<microseconds>\.\d+)?(?P<tzpart>z|[+-]\d{2}:\d{2})?$",
    re.I,
)

ISO_8601_TIME_DURATION_REGEX = re.compile(
    r"^(?P<sign>-?)P(?=\d|T\d)(?:(?P<weeks>\d+)W)?(?:(?P<days>\d+)D)?(?:T(?:(?P<hours>\d+)H)?(?:(?P<minutes>\d+)M)?(?:(?P<seconds>\d+(?:\.\d+)?)S)?)?$",
    re.I,
)


def parse_iso_datetime(value: str) -> datetime:

    if not ISO_8601_DATETIME_REGEX.match(value):
        raise ValueError(f"passed value {value!r} is not valid ISO-8601 datetime.")

    date_parts = ISO_8601_DATETIME_REGEX.findall(value)[0]
    time_part = date_parts[3]
    if ":" in time_part:
        time_part = time_part.split(":")
    else:
        time_part = list(map("".join, zip(*[iter(time_part)] * 2)))

    microseconds_part = date_parts[4]
    microseconds = microseconds_part.lstrip(".") if microseconds_part else "0"
    time_part.append(microseconds)

    if date_parts[5] and date_parts[5].lower() != "z":
        sign = 1 if date_parts[5][0] == "+" else -1
        hours, minutes = date_parts[5][1:].split(":")
        offset = timezone(timedelta(hours=int(hours) * sign, minutes=int(minutes) * sign))
    elif date_parts[5] and date_parts[5].lower() == "z":
        offset = timezone.utc
    else:
        offset = None  # type: ignore

    return datetime(
        year=int(date_parts[0]),
        month=int(date_parts[1]),
        day=int(date_parts[2]),
        hour=int(time_part[0]),
        minute=int(time_part[1]),
        second=int(time_part[2]),
        microsecond=int(time_part[3]),
        tzinfo=offset,
    )


def parse_iso_date(value: str) -> date:
    if not ISO_8601_DATE_REGEX.match(value):
        raise ValueError("Passed value is not valid ISO-8601 date.")

    date_parts = ISO_8601_DATE_REGEX.findall(value)[0]
    return date(year=int(date_parts[0]), month=int(date_parts[1]), day=int(date_parts[2]))


def parse_iso_duration(value: str) -> timedelta:
    """
    Parses duration string according to ISO 8601 and returns timedelta representation (it excludes year and month)
    http://www.datypic.com/sc/xsd/t-xsd_dayTimeDuration.html
    :param str value:
    :return dict:
    """
    if not ISO_8601_TIME_DURATION_REGEX.match(value):
        raise ValueError(f"Passed value {value} is not valid ISO-8601 duration.")

    duration = ISO_8601_TIME_DURATION_REGEX.fullmatch(value)
    sign = -1 if duration.group("sign") else 1  # type: ignore

    kwargs = {
        "weeks": int(duration.group("weeks")) * sign if duration.group("weeks") else 0,  # type: ignore
        "days": int(duration.group("days")) * sign if duration.group("days") else 0,  # type: ignore
        "hours": int(duration.group("hours")) * sign if duration.group("hours") else 0,  # type: ignore
        "minutes": int(duration.group("minutes")) * sign  # type: ignore
        if duration.group("minutes")  # type: ignore
        else 0,
        "seconds": float(duration.group("seconds")) * sign  # type: ignore
        if duration.group("seconds")  # type: ignore
        else 0,
    }

    return timedelta(**kwargs)  # type: ignore


def parse_iso_time(value: str) -> time:
    if not ISO_8601_TIME_REGEX.match(value):
        raise ValueError(f"Passed value {value} is not valid ISO-8601 time.")

    time_parts = ISO_8601_TIME_REGEX.fullmatch(value)
    hour_parts = time_parts.group("time")  # type: ignore
    if ":" in hour_parts:
        hour_parts = hour_parts.split(":")
    else:
        hour_parts = list(map("".join, zip(*[iter(hour_parts)] * 2)))

    microseconds = time_parts.group("microseconds")  # type: ignore
    if microseconds is not None:
        microseconds = int(microseconds[1:])
    else:
        microseconds = 0

    tz_part = time_parts.group("tzpart")  # type: ignore
    if tz_part and tz_part.lower() != "z":
        sign = 1 if tz_part[0] == "+" else -1
        hours, minutes = tz_part[1:].split(":")
        offset = timezone(timedelta(hours=int(hours) * sign, minutes=int(minutes) * sign))
    elif tz_part and tz_part.lower() == "z":
        offset = timezone.utc
    else:
        offset = None  # type: ignore

    return time(
        hour=int(hour_parts[0]),
        minute=int(hour_parts[1]),
        second=int(hour_parts[2]),
        microsecond=microseconds,
        tzinfo=offset,
    )


def timedelta_to_iso_duration(value: timedelta) -> str:
    seconds = value.total_seconds()
    sign = "-" if seconds < 0 else ""
    seconds = abs(seconds)
    minutes, seconds = divmod(seconds, 60)
    hours, minutes = divmod(minutes, 60)
    days, hours = divmod(hours, 24)
    weeks, days = divmod(days, 7)
    weeks, days, hours, minutes = map(int, (weeks, days, hours, minutes))
    seconds = round(seconds, 6)

    iso_8601 = sign + "P"
    iso_8601_date = ""
    iso_8601_time = ""
    if weeks:
        iso_8601_date += f"{weeks}W"

    if days:
        iso_8601_date += f"{days}D"

    if hours:
        iso_8601_time += f"{hours}H"

    if minutes:
        iso_8601_time += f"{minutes}M"

    if seconds:
        if seconds.is_integer():
            iso_8601_time += f"{int(seconds)}S"
        else:
            iso_8601_time += f"{seconds}S"

    return f"{iso_8601}{iso_8601_date}" + (f"T{iso_8601_time}" if iso_8601_time else "")
